from sendbee_api.query_params import QueryParams


class ListTeams(QueryParams):
    """Parameters for list of teams"""

    member_id = 'member_id', 'Fetch teams for a member'


class ListTeamMembers(QueryParams):
    """Parameters for list of team members"""

    team_id = 'team_id', 'Fetch members of one team'



from copy import deepcopy

from sendbee_api.fields import Field, ModelField, NumberField, TextField


class Model:
    """Abstract model class."""

    def __init__(self, item):
        self.item = item
        self.attributes = {}

    def __getattr__(self, item):
        if item in self.attributes.keys():
            attr = self.attributes.get(item)
            if attr:
                return self.attributes.get(item).value
            else:
                return attr
        else:
            raise AttributeError(item)

    @classmethod
    def process(cls, data):
        """Transform raw data into models."""

        model_list = []

        # iterate over formatted data from API call
        for item in data:

            # instantiate model with one row of data
            model_object = cls(item)

            # iterate over model properties
            for model_property in list(cls.__dict__):

                # invoke model property
                property_object = getattr(model_object,
                                          model_property)

                if isinstance(property_object, ModelField):

                    # recursion:
                    # if the instance of the field class is ModeClass
                    # run process on that model with only a portion of data
                    if isinstance(item[property_object.index], list):
                        models = property_object.model_cls.process(
                            item[property_object.index]
                        )
                    elif isinstance(item[property_object.index], dict):
                        models = property_object.model_cls.process(
                            [item[property_object.index]]
                        )[0]

                    setattr(model_object, model_property[1:], models)

                elif isinstance(property_object, Field):

                    # convert raw property data into corespondent
                    # data type defined by the Field class
                    property_object.convert_item(model_object)

                    # set model property value with formatted data
                    if property_object.value is None:
                        model_object.attributes[model_property[1:]] = None
                    else:
                        model_object.attributes[model_property[1:]] = \
                            deepcopy(property_object)

            model_list.append(model_object)

        return model_list


class Meta(Model):

    _total = NumberField(index='total', desc="Total")
    _items_to = NumberField(index='to', desc="Items To")
    _items_from = NumberField(index='from', desc="Items From")
    _per_page = NumberField(index='per_page', desc="Per Page")
    _last_page = NumberField(index='last_page', desc="Last Page")
    _current_page = NumberField(index='current_page', desc="Current Page")


class ServerMessage(Model):

    _message = TextField(index='message', desc='Message from the server')



from sendbee_api.models import Model
from sendbee_api.fields import TextField, BooleanField


class ChatbotActivityStatus(Model):
    """Data model for chatbot status"""

    _conversation_id = TextField(index='conversation_id', desc='Conversation UUID')
    _chatbot_active = BooleanField(index='chatbot_active', desc='Chatbot status')



from sendbee_api.models import Model
from sendbee_api.fields import TextField, BooleanField


class RateLimitError(Model):
    """Data model for rate limit error"""

    _detail = TextField(index='detail', desc='Message detail')
    _error = BooleanField(index='error', desc='Error or not')
    _type = TextField(index='type', desc='Message type')



from sendbee_api.models import Model
from sendbee_api.fields import TextField, DatetimeField, ModelField, ListField


class ContactTag(Model):
    """Data model for contact tags"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')


class ContactField(Model):
    """Data model for contact contact fields"""

    _id = TextField(index='id', desc='Field UUID')
    _type = TextField(index='type', desc='Type')
    _name = TextField(index='name', desc='Name')
    _options = ListField(index='options', desc='List type options')


class Note(Model):
    """Data model for contact notes"""

    _note = TextField(index='note', desc='Note')


class ContactContactField(Model):
    """Data model for contact contact fields"""

    _key = TextField(index='key', desc='Contact field key')
    _value = TextField(index='value', desc='Contact field value')


class ContactNote(Model):
    """Data model for contact notes"""

    _value = TextField(index='value', desc='Note value')


class Contact(Model):
    """Data model for contacts"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')
    _phone = TextField(index='phone', desc='Phone number')
    _created_at = DatetimeField(
        index='created_at', desc='Name', format='%Y-%m-%d %H:%M:%'
    )
    _tags = ModelField(ContactTag, index='tags', desc='Tags')
    _status = TextField(index='status', desc='Subscription status')
    _folder = TextField(index='folder', desc='Conversation folder')
    _contact_fields = ModelField(
        ContactContactField, index='contact_fields', desc='Contact field'
    )
    _notes = ModelField(ContactNote, index='notes', desc='Notes')



from sendbee_api.query_params import QueryParams


class ChatbotActivity(QueryParams):
    """Parameters for turning the chatbot on or off for a conversation"""

    conversation_id = 'conversation_id', 'Conversation UUID'
    active = 'active', 'Chatbot activity'


class ChatbotActivityStatus(QueryParams):
    """Parameters for getting chatbot status for a conversation"""

    conversation_id = 'conversation_id', 'Conversation UUID'



from sendbee_api import constants
from sendbee_api.bind import bind_request
from sendbee_api.models import ServerMessage
from sendbee_api.automation import query_params
from sendbee_api.automation.models import ChatbotActivityStatus


class Automation:
    """Api client for automation"""

    chatbot_activity = bind_request(
        api_path='/automation/chatbot/activity',
        model=ServerMessage,
        query_parameters=query_params.ChatbotActivity,
        method=constants.RequestConst.PUT,
        description='Api client for turning the '
                    'chatbot on or off for a conversation'
    )

    chatbot_activity_status = bind_request(
        api_path='/automation/chatbot/activity/status',
        model=ChatbotActivityStatus,
        force_single_model_response=True,
        query_parameters=query_params.ChatbotActivityStatus,
        method=constants.RequestConst.GET,
        description='Api client for getting current '
                    'chatbot status for a conversation'
    )



from sendbee_api import constants
from sendbee_api.bind import bind_request
from sendbee_api.rate_limit import query_params
from sendbee_api.rate_limit.models import RateLimitError


class RateLimit:
    """Api client for rete limit testing"""

    rate_limit_error_test = bind_request(
        api_path='/rate-limit/error-test',
        model=RateLimitError,
        query_parameters=query_params.RateLimitParams,
        method=constants.RequestConst.GET,
        # ignore_error=True,
        description='Rate limit error test'
    )

    rate_limit_request_test = bind_request(
        api_path='/rate-limit/request-test',
        model=RateLimitError,
        query_parameters=query_params.RateLimitParams,
        method=constants.RequestConst.GET,
        # ignore_error=True,
        description='Rate limit request test'
    )



# Main API class
from sendbee_api.client import Client as SendbeeApi

# Category API subclasses
from sendbee_api.client import Contacts

# Authentication class
from sendbee_api.auth import SendbeeAuth

# Exception classes
from sendbee_api.exceptions import SendbeeRequestApiException, PaginationException



from sendbee_api.teams import models
from sendbee_api.bind import bind_request
from sendbee_api.teams import query_params


class Teams:
    """Api client for conversations"""

    teams = bind_request(
        api_path='/teams/teams',
        model=models.Team,
        query_parameters=query_params.ListTeams,
        description='Api client for teams'
    )
    members = bind_request(
        api_path='/teams/members',
        model=models.Member,
        query_parameters=query_params.ListTeamMembers,
        description='Api client for team members'
    )



from sendbee_api.models import Model
from sendbee_api.fields import TextField, ModelField, BooleanField


class MemberTeam(Model):
    """Data model for member team"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')


class Member(Model):
    """Data model for member"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')
    _role = TextField(index='role', desc='Role')
    _online = TextField(index='online', desc='Online')
    _available = TextField(index='available', desc='Available')
    _teams = ModelField(MemberTeam, index='teams', desc='Teams')


class TeamMember(Model):
    """Data model for team team member"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')
    _role = TextField(index='role', desc='Role')
    _online = TextField(index='online', desc='Online')
    _available = TextField(index='available', desc='Available')


class Team(Model):
    """Data model for team"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')
    _members = ModelField(TeamMember, index='members', desc='Members')



from sendbee_api import constants
from sendbee_api.conversations import models
from sendbee_api.bind import bind_request
from sendbee_api.conversations import query_params


class Messages:
    """Api client for conversations"""

    message_templates = bind_request(
        api_path='/conversations/messages/templates',
        model=models.MessageTemplate,
        query_parameters=query_params.ListMessageTemplates,
        description='Api client for message templates'
    )
    send_template_message = bind_request(
        api_path='/conversations/messages/templates/send',
        model=models.SentMessage,
        method=constants.RequestConst.POST,
        query_parameters=query_params.SendTemplateMessage,
        description='Api client for sending template conversations'
    )
    send_message = bind_request(
        api_path='/conversations/messages/send',
        model=models.SentMessage,
        method=constants.RequestConst.POST,
        query_parameters=query_params.SendMessage,
        description='Api client for sending template conversations'
    )
    conversations = bind_request(
        api_path='/conversations',
        model=models.Conversation,
        query_parameters=query_params.ListConversations,
        description='Api client for fetching conversations'
    )
    messages = bind_request(
        api_path='/conversations/messages',
        model=models.Message,
        query_parameters=query_params.ListMessages,
        description='Api client for fetching conversation messages'
    )

    get_conversation = bind_request(
        api_path='/conversation',
        query_parameters=query_params.GetSingleConversation,
        model=models.Conversation,
        force_single_model_response=True,
        description='Api client for fetching single conversation data'
    )
    update_conversation = bind_request(
        api_path='/conversation',
        method=constants.RequestConst.POST,
        query_parameters=query_params.UpdateSingleConversation,
        model=models.Conversation,
        force_single_model_response=True,
        # query_parameters=query_params.ListMessages,
        description='Api client for updating a conversation'
    )



import ujson

from abc import ABCMeta, abstractmethod

from sendbee_api import constants


class Formatter(metaclass=ABCMeta):
    """Abstract formatter class."""

    def __init__(self, response):
        self._response = response

    @abstractmethod
    def _format_data(self, data):
        """Format method in formatter classes."""

        return data

    @abstractmethod
    def _format_meta(self, data):
        """Format method in formatter classes."""

        return data

    @abstractmethod
    def _format_warning(self, data):
        """Format method in formatter classes."""

        return data

    @abstractmethod
    def _format_headers(self, data):
        """Format method in formatter classes."""

        return data

    def format_data(self, data):
        """Main format method."""

        return self._format_data(data)

    def format_meta(self, data):
        """Main format method."""

        return self._format_meta(data)

    def format_warning(self, data):
        """Main format method."""

        return self._format_warning(data)

    def format_headers(self, data):
        """Main format method."""

        return self._format_headers(data)

    @staticmethod
    def _default_meta_data(total_results):
        """Create default data set if there aren't one from the response."""

        return {
            constants.ResponseDataConst.TOTAL_RESULTS: total_results,
            constants.ResponseDataConst.TOTAL_PAGES: 1,
            constants.ResponseDataConst.CURRENT_PAGE: 1
        }

    @staticmethod
    def _meta_data_to_int(meta_data):
        """Cast all data to int."""

        return {
            constants.ResponseDataConst.TOTAL_RESULTS:
                int(meta_data.get(constants.ResponseDataConst.TOTAL_RESULTS, 1)),
            constants.ResponseDataConst.TOTAL_PAGES:
                int(meta_data.get(constants.ResponseDataConst.TOTAL_PAGES, 1)),
            constants.ResponseDataConst.CURRENT_PAGE:
                int(meta_data.get(constants.ResponseDataConst.CURRENT_PAGE, 1)),
        }


class Json(Formatter):
    """JSON data formatter class."""

    def _unpack_response(self, data):
        """JSON decode"""

        try:
            return ujson.loads(data)
        except ValueError:
            return constants.ResponseConst.DEFAULT_ERROR_MESSAGE

    def _format_data(self, data):
        """Transform raw data from server into python native type."""

        formatted_data = self._unpack_response(data)
        if isinstance(formatted_data, dict):
            return formatted_data.get(
                constants.ResponseDataConst.DATA, formatted_data
            )
        else:
            return formatted_data

    def _format_meta(self, data):
        """Transform raw data from server into python native type."""

        formatted_data = self._unpack_response(data)
        if isinstance(formatted_data, dict):
            return formatted_data.get(constants.ResponseDataConst.META, {})
        else:
            return self.__class__._default_meta_data(len(formatted_data))

    def _format_warning(self, data):
        """Transform raw warning data from server into python native type."""

        formatted_data = self._unpack_response(data)
        if isinstance(formatted_data, dict):
            return formatted_data.get(constants.WarningConst.WARNING)

    def _format_headers(self, data):
        """Transform raw data from server into python native type."""

        return data


class FormatterFactory:
    """Formatter factory class."""

    formatters = {
        constants.FormatterConst.JSON: Json,
    }

    def __init__(self, name: str):
        self._name = name

    def get_formatter(self):
        """Get the formatter class using string key."""

        return FormatterFactory.formatters.get(self._name)



from sendbee_api.query_params import QueryParams


class ListMessageTemplates(QueryParams):
    """Parameters for list of message templates API request"""

    status = 'status', 'Fetch templates with desired status'
    search_query = 'search_query', 'Filter contacts by query string'
    page = 'page', 'Page number for pagination'
    limit = 'limit', 'Number of items per page. Maximum is 100.'


class SendTemplateMessage(QueryParams):
    """Parameters for sending template message"""

    tags = 'tags', 'Template message tags'
    button_tags = 'button_tags', 'Template button tags'
    phone = 'phone', 'Contact\'s phone number'
    language = 'language', 'Template message language'
    template_keyword = 'template_keyword', 'Template message keyword'
    prevent_bot_off = 'prevent_bot_off', 'Prevent turning-off chatbot'
    agent_id = 'agent_id', 'Assigned agent for the conversation'
    attachment = 'attachment', 'Attachment URL for media templates'


class SendMessage(QueryParams):
    """Parameters for sending message"""

    phone = 'phone', 'Contact\'s phone number'
    text = 'text', 'Message text'
    media_url = 'media_url', 'Media URL for media message'
    prevent_bot_off = 'prevent_bot_off', 'Prevent turning-off chatbot'
    agent_id = 'agent_id', 'Assigned agent for the conversation'


class ListConversations(QueryParams):
    """Parameters for list of conversations"""

    date_from = 'date_from', 'Fetch all conversations from this date'
    date_to = 'date_to', 'Fetch all conversations to this date'
    folder = 'folder', 'Filter conversation by folder'
    search_query = 'search_query', 'Filter conversations by query string'
    page = 'page', 'Page number for pagination'
    limit = 'limit', 'Number of items per page. Maximum is 100.'


class ListMessages(QueryParams):
    """Parameters for list of messages"""

    conversation_id = 'conversation_id', 'Conversation UUID'
    page = 'page', 'Page number for pagination'


class GetSingleConversation(QueryParams):
    """Parameters for getting a single conversation"""
    conversation_id = 'conversation_id', 'Conversation UUID'


class UpdateSingleConversation(QueryParams):
    """Parameters for updating a single conversation"""
    conversation_id = 'conversation_id', 'Conversation UUID'
    folder = 'folder', 'Assign conversation to specific folder, open|done'



from collections import namedtuple



RequestConst = namedtuple(
    'RequestConst', [
        'GET', 'POST', 'PUT', 'DELETE', 'PATH', 'QUERY', 'METHOD', 'TIMEOUT',
        'API_PATH', 'PROTOCOL', 'QUERY_PARAMETERS', 'DEFAULT_PARAMETERS',
        'BLOCK_NOTIFICATIONS', 'BLOCK_AUTOMATION'
    ]
)(
    'GET', 'POST', 'PUT', 'DELETE', 'path', 'query', 'method', 'timeout',
    'api_path', 'protocol', 'query_parameters', 'default_parameters',
    'block_notifications', 'block_automation'
)

ErrorConst = namedtuple(
    'ErrorConst', ['ERROR', 'UNRECOGNIZED_ERROR', 'DETAIL', 'NOT_FOUND']
)('error', 'Unrecognized error', 'detail', 'Endpoint not found')

WarningConst = namedtuple(
    'WarningConst', ['WARNING', 'MESSAGE']
)('warning', '[Sendbee API] WARNING: ')

ResponseConst = namedtuple(
    'ResponseConst', [
        'RESPONSE', 'STATUS_CODE', 'RESPONSE_OBJECT', 'DEFAULT_ERROR_MESSAGE'
    ]
)(
    'response', 'status_code', 'response_object',
    {
        ErrorConst.ERROR: {
            ErrorConst.DETAIL: ErrorConst.UNRECOGNIZED_ERROR,
            'type': 'unrecognized_error'
        }
    }
)

ClientConst = namedtuple(
    'ClientConst', [
        'META', 'MODEL', 'MODELS', 'FORMATTER', 'DESCRIPTION',
        'FORCE_SINGLE_MODEL_RESPONSE', 'IGNORE_ERROR', 'HEADERS'
    ]
)('meta', 'model', 'models', 'formatter', 'description',
  'force_single_model_response', 'ignore_error', 'headers')

FormatterConst = namedtuple(
    'FormatterConst', ['JSON', 'FORMATTED']
)('json', 'formatted')

TestConst = namedtuple('TestConst', ['FAKE_RESPONSE_PATH'])('fake_response_path')

ResponseCode = namedtuple(
    'ResponseCode', ['OK', 'NOT_FOUND', 'BAD_REQUEST']
)(200, 404, 400)

BoolConst = namedtuple('BoolConst', ['TRUE', 'FALSE'])('1', '0')

MiscConst = namedtuple(
    'MiscConst', ['FORMAT', 'PRINT_PARAMS']
)('format', 'print_params')

ResponseDataConst = namedtuple(
    'ResponseDataConst', ['DATA', 'META']
)('data', 'meta')




from sendbee_api import constants
from sendbee_api.contacts import models
from sendbee_api.bind import bind_request
from sendbee_api.models import ServerMessage
from sendbee_api.contacts import query_params


class Contacts:
    """Api client for contacts"""

    contacts = bind_request(
        api_path='/contacts',
        model=models.Contact,
        query_parameters=query_params.ListContacts,
        description='Api client for contacts'
    )
    subscribe_contact = bind_request(
        api_path='/contacts/subscribe',
        model=models.Contact,
        method=constants.RequestConst.POST,
        query_parameters=query_params.UpdateContacts,
        default_parameters={
            constants.RequestConst.BLOCK_AUTOMATION: True,
            constants.RequestConst.BLOCK_NOTIFICATIONS: True,
        },
        description='Api client for subscribing a contacts'
    )
    update_contact = bind_request(
        api_path='/contacts',
        model=models.Contact,
        method=constants.RequestConst.PUT,
        query_parameters=query_params.UpdateContacts,
        description='Api client for updating contacts'
    )
    tags = bind_request(
        api_path='/contacts/tags',
        model=models.ContactTag,
        query_parameters=query_params.ListTags,
        description='Api client for tags'
    )
    update_tag = bind_request(
        api_path='/contacts/tags',
        model=models.ContactTag,
        method=constants.RequestConst.PUT,
        query_parameters=query_params.UpdateTag,
        description='Api client for updating tags'
    )
    create_tag = bind_request(
        api_path='/contacts/tags',
        model=models.ContactTag,
        method=constants.RequestConst.POST,
        query_parameters=query_params.ListTags,
        description='Api client for updating tags'
    )
    delete_tag = bind_request(
        api_path='/contacts/tags',
        model=ServerMessage,
        method=constants.RequestConst.DELETE,
        query_parameters=query_params.DeleteTag,
        description='Api client for deleting tags'
    )
    contact_fields = bind_request(
        api_path='/contacts/fields',
        model=models.ContactField,
        query_parameters=query_params.ListContactFields,
        description='Api client for contact fields'
    )
    create_contact_field = bind_request(
        api_path='/contacts/fields',
        model=models.ContactField,
        method=constants.RequestConst.POST,
        query_parameters=query_params.CreateContactFields,
        description='Api client for creating contact contact fields'
    )
    update_contact_field = bind_request(
        api_path='/contacts/fields',
        model=models.ContactField,
        method=constants.RequestConst.PUT,
        query_parameters=query_params.UpdateContactFields,
        description='Api client for updating contact contact fields'
    )
    delete_contact_field = bind_request(
        api_path='/contacts/fields',
        model=ServerMessage,
        method=constants.RequestConst.DELETE,
        query_parameters=query_params.DeleteContactFields,
        description='Api client for deleting contact contact fields'
    )



from datetime import datetime

from sendbee_api import constants


class Field:
    """Abstract field class."""

    def __init__(self, index, desc=None, **kwargs):
        self.index = index
        self._kwargs = kwargs

        self.__doc__ = desc
        self.value = None

    def convert_item(self, model):
        """Convert item to desired item type"""

        try:
            self.value = model.item[self.index]
        except IndexError:
            self.value = None
        except KeyError:
            self.value = None

        try:
            if self.value is not None:
                self.value = self._convert_field_item(
                    self.value, **self._kwargs
                )
        except TypeError:
            self.value = None

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        return data


class NumberField(Field):
    """Converting item to integer."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        try:
            return int(data)
        except ValueError:
            return 0


class RealNumberField(Field):
    """Converting item to integer."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        try:
            return float(data)
        except ValueError:
            return 0.0


class TextField(Field):
    """Converting item to string."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        try:
            return str(data)
        except ValueError:
            return ''


class BooleanField(Field):
    """Converting item to boolean."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        return bool(data)


class DatetimeField(Field):
    """Converting item to datetime object."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        try:
            _format = kwargs.get(constants.MiscConst.FORMAT)
            return datetime.strptime(data, _format)
        except ValueError:
            return data


class ListField(Field):
    """Converting item to list."""

    def _convert_field_item(self, data, **kwargs):
        """Actual converting."""

        try:
            return list(data)
        except ValueError:
            return data


class ModelField(Field):
    """Converting item to another data model."""

    def __init__(self, model_cls, index, desc=None, **kwargs):
        self.model_cls = model_cls
        super().__init__(index, desc, **kwargs)



import hmac
import base64
import hashlib
from datetime import datetime, timezone, timedelta


class SendbeeAuth:
    """Authentication class for Sendbee API"""

    def __init__(self, private_key):
        if isinstance(private_key, str):
            private_key = private_key.encode('utf-8')
        self._private_key = private_key

    def _get_encrypted_key(self, timestamp):
        """Generates encrypted key from timestamp and private key
        :param timestamp: timestamp string
        :return: key string
        """

        if isinstance(timestamp, str):
            timestamp = timestamp.encode('utf-8')

        return hmac.new(
            self._private_key, timestamp, hashlib.sha256
        ).hexdigest()

    def get_auth_token(self):
        """Generates auth token from timestamp and encrypted key
        :return: token string
        """

        timestamp = str(int(
            datetime.now(timezone.utc).timestamp()
        ))
        encrypted = self._get_encrypted_key(timestamp)
        ts_encrypt = f'{timestamp}.{encrypted}'.encode('utf-8')

        return base64.b64encode(ts_encrypt).decode('utf-8')

    def check_auth_token(self, token, expiration_seconds=60*15):
        """Checks if the provided and generated tokens are equal
        :param token: token string
        :param expiration_seconds: seconds integer
        :return: True or False
        """

        if isinstance(token, str):
            token = token.encode('utf-8')

        timestamp, encrypted = \
            base64.b64decode(token).decode('utf-8').split('.')

        if datetime.fromtimestamp(int(timestamp), tz=timezone.utc) > \
                datetime.now(timezone.utc)+timedelta(
                    minutes=int(expiration_seconds)):
            return False

        if encrypted != self._get_encrypted_key(timestamp):
            return False

        return True



class SendbeeException(Exception):
	"""Handle All Sendbee Exceptions"""

	def __init__(self, *args, **kwargs):
		self.response = kwargs.get('response')


class SendbeeRequestApiException(SendbeeException):
	"""Handle Request Exceptions"""


class SendbeeFormatterException(SendbeeException):
	"""Handle Formatter Exceptions"""


class PaginationException(SendbeeException):
	"""Handle Pagination Exceptions"""



from sendbee_api.models import Model
from sendbee_api.fields import TextField, ModelField, BooleanField, \
    DatetimeField, NumberField, RealNumberField


class TemplateTag(Model):
    """Data model for template tags"""

    _name = TextField(index='name', desc='Name')


class TemplateButton(Model):
    """Data model for template button"""

    _index = NumberField(index='index', desc='Index')
    _type = TextField(index='type', desc='Type')
    _title = TextField(index='title', desc='Title')
    _value = TextField(index='value', desc='Value')


class MessageTemplate(Model):
    """Data model for message templates"""

    _id = TextField(index='id', desc='UUID')
    _status = TextField(index='status', desc='Status')
    _keyword = TextField(index='keyword', desc='Keyword')
    _tags = ModelField(TemplateTag, index='tags', desc='Tags')
    _button_tags = ModelField(
        TemplateTag, index='button_tags', desc='Button tags')
    _text = TextField(index='text', desc='Text')
    _buttons = ModelField(TemplateButton, index='buttons', desc='Buttons')
    _language = TextField(index='language', desc='Language')
    _attachment = TextField(index='attachment', desc='Attachment')
    _rejected_reason = TextField(
        index='rejected_reason',
        desc='Reason why Facebook didn\'t approve this template')


class SentMessage(Model):
    """Data model for sent message"""

    _conversation_id = TextField(index='id', desc='Conversation UUID')
    _message_id = TextField(index='message_id', desc='Message UUID')
    _status = TextField(index='status', desc='Message status')


class ConversationContact(Model):
    """Data model for contact in conversation"""

    _id = TextField(index='id', desc='UUID')
    _name = TextField(index='name', desc='Name')
    _phone = TextField(index='phone', desc='Phone number')


class ConversationLastMessage(Model):
    """Data model for last message in conversation"""

    _direction = TextField(index='direction', desc='Last message direction')
    _status = TextField(index='status', desc='Last message status')
    _inbound_sent_at = DatetimeField(
        index='inbound_sent_at', desc='Last inbound message sent at',
        format='%Y-%m-%d %H:%M:%'
    )
    _outbound_sent_at = DatetimeField(
        index='outbound_sent_at', desc='Last outbound message sent at',
        format='%Y-%m-%d %H:%M:%'
    )


class Conversation(Model):
    """Data model for conversations"""

    _id = TextField(index='id', desc='UUID')
    _folder = TextField(index='folder', desc='Folder')
    _last_message = ModelField(
        ConversationLastMessage, index='last_message', desc='Last message'
    )
    _contact = ModelField(ConversationContact, index='contact', desc='Contact')
    _chatbot_active = BooleanField(index='chatbot_active', desc='Approved')
    _platform = TextField(index='platform', desc='Platform')
    _created_at = DatetimeField(
        index='created_at', desc='Created at', format='%Y-%m-%d %H:%M:%'
    )


class Message(Model):
    """Data model for messages"""

    _body = TextField(index='body', desc='Message body')
    _media_type = TextField(index='media_type', desc='Message media type')
    _media_url = TextField(index='media_url', desc='Media URL')
    _status = TextField(index='status', desc='Message status')
    _direction = TextField(index='direction', desc='Message direction')
    _sent_at = RealNumberField(index='sent_at', desc='Message sent at')



from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='sendbee_api',
    version='1.7.0',

    description='Python client SDK for Sendbee Public API',
    long_description=readme(),
    long_description_content_type='text/markdown',

    url='https://github.com/sendbee/sendbee-python-client',
    licence='MIT',

    author='Sendbee ltd',
    author_email='info@sendbee.io',

    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='sendbee api python',

    packages=find_packages(),
    install_requires=[
        'click>=7.0',
        'requests>=2.20.0',
        'dumpit>=0.5.0',
        'aenum>=2.1.2',
        'ujson==2.0.1',
        'cryptography>=3.2'
    ],

    project_urls={
        'Source': 'https://github.com/sendbee/sendbee-python-client',
    },
)



import click
from dumpit import pdumpit
from aenum import MultiValueEnum


class DefaultQueryParams(MultiValueEnum):
    """Default set of query parameters."""

    msg_type = 'msgtype', 'simple or extended (extended cost more credits)'
    protocol = 'protocol', 'Response type. Use one of the following: '\
                           'xml, csv, json, jsono (object)'


class QueryParams(MultiValueEnum):
    """Query parameters for API calls."""

    @classmethod
    def get_params(cls):
        """Merges default query parameters with
        call specific query parameters."""

        return {**{param.name: param.value for param in cls},
                **{param.name: param.value for param in DefaultQueryParams}}

    @classmethod
    def print_params(cls):
        """Prints all query parameters."""

        for param in DefaultQueryParams:
            param.__doc__ = param.values[1]

        click.echo('Default params:')
        pdumpit(DefaultQueryParams, all_=False)

        for param in cls:
            param.__doc__ = param.values[1]

        click.echo('{} params:'.format(cls.__name__))
        pdumpit(cls, all_=False)



from sendbee_api import constants


class Response:
    """Response object returned from API call."""

    def __init__(self, data, headers, status_code, formatter, api_request):
        self._data = data
        self._headers = headers
        self._model = api_request.model

        self.api_reguest = api_request
        self.formatter = formatter(self)
        self.status_code = status_code

        self._response_data = {
            constants.FormatterConst.FORMATTED: None,
            constants.ClientConst.MODELS: None,
            constants.ClientConst.META: None,
            constants.WarningConst.WARNING: None,
            constants.ClientConst.HEADERS: None,
        }

    def __iter__(self):
        return iter(self.models)

    @property
    def raw_data(self):
        """Return data as is from the server."""

        return self._data

    @property
    def formatted_data(self):
        """Format data using selected formatter."""

        if self._response_data[constants.FormatterConst.FORMATTED] is None:
            self._response_data[constants.FormatterConst.FORMATTED] = \
                self.formatter.format_data(self._data)

        return self._response_data[constants.FormatterConst.FORMATTED]

    @property
    def models(self):
        """Transform raw data into models."""

        if not self._response_data[constants.ClientConst.MODELS]:
            formatted_data = self.formatter.format_data(self._data)
            if not isinstance(formatted_data, list):
                formatted_data = [formatted_data]
            self._response_data[constants.ClientConst.MODELS] = \
                self._model.process(formatted_data)

        return self._response_data[constants.ClientConst.MODELS]

    @property
    def meta(self):
        """Transform raw meta data into models."""

        if not self._response_data[constants.ClientConst.META]:
            from sendbee_api.models import Meta
            formatted_data = self.formatter.format_meta(self._data)
            self._response_data[constants.ClientConst.META] = \
                Meta.process([formatted_data])[0]

        return self._response_data[constants.ClientConst.META]

    @property
    def headers(self):
        """Response headers."""

        if not self._response_data[constants.ClientConst.HEADERS]:
            self._response_data[constants.ClientConst.HEADERS] = \
                self.formatter.format_headers(self._headers)

        return self._response_data[constants.ClientConst.HEADERS]

    @property
    def warning(self):
        """Format warning data using selected formatter."""

        if self._response_data[constants.WarningConst.WARNING] is None:
            self._response_data[constants.WarningConst.WARNING] = \
                self.formatter.format_warning(self._data)

        return self._response_data[constants.WarningConst.WARNING]

    def has_next(self):
        """Does endpoint has another page to fetch."""

        return self.meta.current_page < self.meta.last_page

    def next_page(self):
        """Get next page number."""

        return self.meta.current_page + 1



from sendbee_api.constants import RequestConst
from sendbee_api.query_params import QueryParams


class ListContacts(QueryParams):
    """Get the list of business' contacts"""

    tags = 'tags', 'Filter contacts by tags'
    status = 'status', 'Filter contacts by status'
    search_query = 'search_query', 'Filter contacts by query string'
    page = 'page', 'Page number for pagination'
    limit = 'limit', 'Number of items per page. Maximum is 100.'


class UpdateContacts(QueryParams):
    """Update a contact"""

    id = 'id', 'Contact ID'
    phone = 'phone', 'Contact phone number'
    tags = 'tags', 'Tag contact while subscribing it'
    name = 'name', 'Subscriber name'
    contact_fields = 'contact_fields', 'Contact data fields'
    notes = 'notes', 'Notes about subscriber'
    block_notifications = RequestConst.BLOCK_NOTIFICATIONS, \
                          'Block sending notifications ' \
                          'after subscribing contact'


class ListTags(QueryParams):
    """Get the list of contact tags"""

    name = 'name', 'Name of the tag'
    page = 'page', 'Page number for pagination'
    limit = 'limit', 'Number of items per page. Maximum is 100.'


class UpdateTag(QueryParams):
    """Update tag"""

    id = 'id', 'ID of the tag'
    name = 'name', 'Name of the tag'


class DeleteTag(QueryParams):
    """Delete tag"""

    id = 'id', 'ID of the tag'


class ListContactFields(QueryParams):
    """Get the list of contact contact fields"""

    search_query = 'search_query', 'Filter contact contact ' \
                                   'fields by query string'
    page = 'page', 'Page number for pagination'
    limit = 'limit', 'Number of items per page. Maximum is 100.'


class CreateContactFields(QueryParams):
    """Create contact contact fields"""

    name = 'name', 'Name of the contact field'
    type = 'type', 'Type of the contact field'
    options = 'options', 'List type options'


class UpdateContactFields(QueryParams):
    """Update contact contact fields"""

    id = 'id', 'Contact field UUID'
    name = 'name', 'Name of the contact field'
    type = 'type', 'Type of the contact field'
    options = 'options', 'List type options'


class DeleteContactFields(QueryParams):
    """Delete tag"""

    id = 'id', 'Contact field UUID'



import click
import requests
from abc import ABCMeta

from sendbee_api import constants
from sendbee_api.debug import Debug
from sendbee_api.auth import SendbeeAuth
from sendbee_api.response import Response
from sendbee_api.formatter import FormatterFactory
from sendbee_api.exceptions import SendbeeRequestApiException, PaginationException


class Api(metaclass=ABCMeta):
    """Abstract API class."""


def bind_request(**request_data):
    """Binds request class to client property, dynamically."""

    class Request(Api):
        """Request class. Does the actual API request."""

        model = request_data.get(constants.ClientConst.MODEL)
        api_path = request_data.get(constants.RequestConst.API_PATH)
        formatter = request_data.get(constants.ClientConst.FORMATTER)
        ignore_error = request_data.get(
            constants.ClientConst.IGNORE_ERROR, False
        )
        method = request_data.get(
            constants.RequestConst.METHOD, constants.RequestConst.GET
        )
        query_parameters = request_data.get(
            constants.RequestConst.QUERY_PARAMETERS
        )
        fake_response_path = request_data.get(
            constants.TestConst.FAKE_RESPONSE_PATH
        )
        default_parameters = request_data.get(
            constants.RequestConst.DEFAULT_PARAMETERS, {}
        )
        # immediately determine type of response
        single_model_response = request_data.get(
            constants.ClientConst.FORCE_SINGLE_MODEL_RESPONSE, False
        ) or method in [
            constants.RequestConst.POST,
            constants.RequestConst.PUT,
            constants.RequestConst.DELETE
        ]

        def __init__(self, client, debug: 'Debug',
                     *path_params, **query_params):
            client.request = self

            self.url = None
            self.debug = debug
            self.client = client
            self.parameters = {
                constants.RequestConst.QUERY: {},
                constants.RequestConst.PATH: []
            }

            self._timeout = 20

            self._set_parameters(*path_params, **query_params)

        def _set_parameters(self, *path_params, **query_params):
            """
            Prepares the list of query parameters
            :path_params: list of path parameters
            :query_params: dict of query parameters
            :return: None
            """

            # take timeout
            try:
                self._timeout = int(query_params.get(
                    constants.RequestConst.TIMEOUT, self._timeout
                ))
            except ValueError:
                pass
            try:
                del query_params[constants.RequestConst.TIMEOUT]
            except KeyError:
                pass

            # set default API call params
            for key, value in self.default_parameters.items():
                self.parameters[constants.RequestConst.QUERY][key] = value

            _query_params = self.query_parameters.get_params()

            # set API call params defined during the "call" invocation
            for key, value in query_params.items():
                if value is None:
                    continue

                if key in _query_params.values():
                    self.parameters[constants.RequestConst.QUERY][key] = value

                elif key in _query_params.keys():
                    self.parameters[
                        constants.RequestConst.QUERY
                    ][_query_params[key]] = value

            if self.method == constants.RequestConst.GET:
                # transform all True and False param to 1 and 0
                for key, value in self.parameters[
                    constants.RequestConst.QUERY
                ].items():
                    if value is True:
                        self.parameters[constants.RequestConst.QUERY][key] = \
                            constants.BoolConst.TRUE
                    if value is False:
                        self.parameters[constants.RequestConst.QUERY][key] = \
                            constants.BoolConst.FALSE

            # set optional url path params
            for value in path_params:
                self.parameters[constants.RequestConst.PATH].append(value)

        def _prepare_url(self):
            """
            Prepares url and query parameters for the request
            :return: URL
            """

            base_url = '{}://{}{}'.format(
                self.client.protocol, self.client.base_url, self.api_path
            )
            url_parts = '/'.join(
                [part for part in self.parameters[constants.RequestConst.PATH]]
            )

            if url_parts:
                final_url = '{}/{}'.format(base_url, url_parts)
            else:
                final_url = base_url

            if self.method == constants.RequestConst.GET:
                params = self.parameters[constants.RequestConst.QUERY]
                for param, value in params.items():
                    if isinstance(value, list):
                        params[param] = ','.join(value)
                    elif isinstance(value, dict):
                        params[param] = ','.join([f'{k}:{v}' for k, v in value])

                url_query = '?' + '&'.join([f'{k}={v}' for k, v in params.items()])
                final_url = '{}{}'.format(final_url, url_query)

            self.debug.ok('final url', final_url)

            return final_url

        def _headers(self):
            """Construct headers data with authentication part"""

            auth_token = SendbeeAuth(self.client.api_secret).get_auth_token()
            headers = {
                'X-Auth-Token': auth_token,
                'X-Api-Key': self.client.api_key,
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'User-Agent': 'Sendbee Python API Client'
            }
            self.debug.ok('headers', headers)

            return headers

        def _do_request(self, url: str):
            """
            Makes the request to Sendbee Api servers
            :url: Url for the request
            :return: Tuple with two elements, status code and content
            """

            self.debug.ok('method', self.method)

            if self.client.fake_response_path:
                with open(self.client.fake_response_path, 'r') as f:
                    return constants.ResponseCode.OK, f.read()

            elif self.method == constants.RequestConst.GET:
                response = requests.get(
                    url, headers=self._headers(), timeout=self._timeout
                )
                response.encoding = 'utf-8'

                self.debug.ok(
                    constants.RequestConst.QUERY_PARAMETERS,
                    self.parameters[constants.RequestConst.QUERY]
                )
                self.debug.ok(constants.ResponseConst.RESPONSE_OBJECT, response)

                return response.status_code, response.headers, response.text

            elif self.method in [
                constants.RequestConst.POST,
                constants.RequestConst.PUT,
                constants.RequestConst.DELETE
            ]:
                if self.method == constants.RequestConst.POST:
                    send_request = requests.post
                elif self.method == constants.RequestConst.PUT:
                    send_request = requests.put
                elif self.method == constants.RequestConst.DELETE:
                    send_request = requests.delete

                response = send_request(
                    url, json=self.parameters[constants.RequestConst.QUERY],
                    headers=self._headers(), timeout=self._timeout
                )

                self.debug.ok('payload', self.parameters[
                    constants.RequestConst.QUERY
                ])
                self.debug.ok(constants.ResponseConst.RESPONSE_OBJECT, response)

                return response.status_code, response.headers, response.text

            else:
                return constants.ResponseCode.NOT_FOUND, {}

        def _process_response(self, status_code, response, headers):
            """
            Process response using models
            :status_code: Response status code
            :response: Content
            :return: Response object
            """

            formatter = self.formatter
            if not formatter:
                formatter = FormatterFactory(constants.FormatterConst.JSON)\
                    .get_formatter()

            response = Response(response, headers, status_code, formatter, self)
            formatted_data = response.formatted_data

            if status_code >= constants.ResponseCode.BAD_REQUEST and \
                    self.ignore_error is False:

                if status_code == constants.ResponseCode.NOT_FOUND:
                    error_msg = \
                        constants.ErrorConst.NOT_FOUND
                elif constants.ErrorConst.ERROR not in formatted_data:
                    error_msg = \
                        constants.ResponseConst.DEFAULT_ERROR_MESSAGE
                else:
                    try:
                        error_msg = formatted_data.get(
                            constants.ErrorConst.ERROR, {}
                        ).get(
                            constants.ErrorConst.DETAIL,
                            constants.ErrorConst.UNRECOGNIZED_ERROR
                        )
                    except AttributeError:
                        error_msg = formatted_data.get(
                            constants.ErrorConst.DETAIL,
                            constants.ErrorConst.UNRECOGNIZED_ERROR
                        )

                self.debug.error(
                    constants.ResponseConst.STATUS_CODE, status_code
                )
                self.debug.error(
                    constants.ResponseConst.RESPONSE, response.formatted_data
                )
                raise SendbeeRequestApiException(error_msg, response=response)
            else:
                self.debug.ok(constants.ResponseConst.STATUS_CODE, status_code)
                self.debug.ok(constants.ResponseConst.RESPONSE, response.raw_data)

                if response.meta.current_page:
                    if response.meta.current_page > 1 and len(response.models) == 0:
                        raise PaginationException(
                            f'Page {response.meta.current_page} has no data'
                        )

            if response.warning:
                click.secho(
                    constants.WarningConst.MESSAGE + response.warning,
                    fg='yellow'
                )

            if self.single_model_response:
                if response.models:
                    return response.models[0]
                else:
                    return None
            else:
                return response

        def call(self):
            """
            Makes the API call
            :return: Return value from self._process_response()
            """

            self.url = self._prepare_url()
            status_code, headers, response = self._do_request(self.url)
            return self._process_response(status_code, response, headers)

    def call(client, *path_params, **query_params):
        """
        Binded method for API calls
        :path_params: list of path parameters
        :query_params: dict of query parameters
        :return: Return value from Request.call()
        """

        if constants.MiscConst.PRINT_PARAMS in query_params:
            Request.query_parameters.print_params()
            return

        with Debug(client=client) as debug:
            request = Request(client, debug, *path_params, **query_params)
            return request.call()

    call.__doc__ = request_data.get(constants.ClientConst.DESCRIPTION)

    return call



import click

from sendbee_api.auth import SendbeeAuth
from sendbee_api.teams.client import Teams
from sendbee_api.contacts.client import Contacts
from sendbee_api.rate_limit.client import RateLimit
from sendbee_api.automation.client import Automation
from sendbee_api.conversations.client import Messages
from sendbee_api.exceptions import SendbeeRequestApiException


class Client(Contacts, Messages, Automation, Teams, RateLimit):
    """Main API class. Sets all API calls."""

    base_url = 'api-v2.sendbee.io'
    protocol = 'https'

    def __init__(self, api_key, api_secret,
                 debug=False, fake_response_path=None):

        if not api_key:
            raise SendbeeRequestApiException('API key missing!')
        if not api_secret:
            raise SendbeeRequestApiException('API secret missing!')

        self.debug = debug
        self.request = None
        self.api_key = api_key
        self.api_secret = api_secret
        self.fake_response_path = fake_response_path

    @classmethod
    def print_params_for(cls, fn_name) -> None:
        """Prints parameters for certain API call function."""

        try:
            getattr(cls, fn_name)(None, print_params=True)
        except AttributeError:
            click.secho('Unknown API method: {}'.format(fn_name), fg='red')

    @property
    def auth(self):
        return SendbeeAuth(self.api_secret)
